"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CacheModeEnum = void 0;
const util_1 = require("util");
const child_process_1 = require("child_process");
const promises_1 = require("fs/promises");
const fs_1 = require("fs");
const path_1 = require("path");
const utilities_1 = require("./utilities");
const tape_1 = require("./tape");
const os_1 = require("os");
const form_data_1 = __importDefault(require("form-data"));
exports.CacheModeEnum = {
    OFF: 0,
    STANDARD: 1,
    AGGRESSIVE: 2
};
class Groove {
    constructor(config) {
        this.process = null;
        this.executablePath = null;
        this.commandTimeout = config.commandTimeout || 5000;
        this.port = config.port || 6010;
        this.controlPort = config.controlPort || 6011;
        this.proxyServer = config.proxyServer || null;
        this.proxyUsername = config.proxyUsername || null;
        this.proxyPassword = config.proxyPassword || null;
        this.authUsername = config.authUsername || null;
        this.authPassword = config.authPassword || null;
        this.silenceLogs = false;
        this.baseUrlProxy = `http://localhost:${this.port}`;
        this.baseUrlControl = `http://localhost:${this.controlPort}`;
        this.certificate = (0, fs_1.readFileSync)((0, path_1.join)((0, os_1.homedir)(), '.grooveproxy/ca.crt'));
    }
    launch() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.process != null) {
                throw Error("Already spawned proxy.");
            }
            const parameters = {
                "--port": this.port ? this.port.toString() : null,
                "--control-port": this.controlPort ? this.controlPort.toString() : null,
                "--proxy-server": this.proxyServer,
                "--proxy-username": this.proxyUsername,
                "--proxy-password": this.proxyPassword,
                "--auth-username": this.authUsername,
                "--auth-password": this.authPassword,
            };
            const exc = yield this.getExecutablePath();
            this.process = (0, child_process_1.spawn)(exc, Object.entries(parameters).reduce((previous, [key, value]) => {
                if (value != null) {
                    previous = [...previous, key, value];
                }
                return previous;
            }, []));
            this.process.stdout.setEncoding('utf8');
            this.process.stdout.on('data', (data) => {
                if (this.silenceLogs)
                    return;
                console.log(`proxy-stdout: ${data}`);
            });
            this.process.stderr.setEncoding('utf8');
            this.process.stderr.on('data', (data) => {
                if (this.silenceLogs)
                    return;
                console.log(`proxy-stderr: ${data}`);
            });
            this.process.on('close', (code) => {
                if (this.silenceLogs)
                    return;
                console.log(`child process exited with code ${code}`);
            });
            // Ensure we cleanup after ourselves
            process.on('exit', () => {
                console.log("Will quit");
                this.stop();
            });
            yield (0, utilities_1.sleep)(1000);
        });
    }
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            this.silenceLogs = true;
            if (this.process)
                this.process.kill();
        });
    }
    tapeStart() {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/tape/record`, {
                method: "POST",
                timeout: this.commandTimeout,
            });
            const contents = yield response.json();
            if (contents["success"] != true) {
                throw Error("Failed to start recording.");
            }
        });
    }
    tapeGet() {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/tape/retrieve`, {
                method: "POST",
                timeout: this.commandTimeout,
            });
            const contents = response.body;
            const session = new tape_1.TapeSession();
            session.readFromServer(yield (0, utilities_1.streamToBuffer)(contents));
            return session;
        });
    }
    tapeLoad(session) {
        return __awaiter(this, void 0, void 0, function* () {
            const formData = new form_data_1.default();
            // Filename is irrelevant but required to upload payload buffer as file
            formData.append("file", session.toServer(), { filename: 'tape.json.gzip' });
            const response = yield (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/tape/load`, {
                method: "POST",
                timeout: this.commandTimeout,
                body: formData,
            });
            const contents = yield response.json();
            if (contents["success"] != true) {
                throw Error("Failed to load tape.");
            }
        });
    }
    tapeStop() {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/tape/stop`, {
                method: "POST",
                timeout: this.commandTimeout,
            });
            const contents = yield response.json();
            if (contents["success"] != true) {
                throw Error("Failed to stop recording.");
            }
        });
    }
    setCacheMode(mode) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield (0, utilities_1.fetchWithTimeout)(`${this.baseUrlControl}/api/cache/mode`, {
                method: "POST",
                timeout: this.commandTimeout,
                body: JSON.stringify({ mode }),
            });
            const contents = yield response.json();
            if (contents["success"] != true) {
                throw Error("Failed to set cache mode.");
            }
        });
    }
    getExecutablePath() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.executablePath)
                return this.executablePath;
            const npmBin = yield (0, util_1.promisify)(child_process_1.exec)("npm bin");
            if (npmBin.stderr || !npmBin.stdout) {
                throw Error("Unknown grooveproxy executable location");
            }
            const binDirectory = npmBin.stdout.trim();
            this.executablePath = (0, path_1.join)(binDirectory, "grooveproxy");
            // Determine if the path exists, will raise an error if not
            yield (0, promises_1.stat)(this.executablePath);
            return this.executablePath;
        });
    }
}
exports.default = Groove;
